/*
 * Copyright © 2007 Jerome Glisse <glisse@freedesktop.org>
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <expat.h>
#include "dump.h"

struct xml_parser {
    struct dump *dump;
};

static long count_line = 0;

static void start_func(void *data, const char *el, const char **attr)
{
    struct xml_parser *pdata = (struct xml_parser *)data;
    struct mmreg *mmreg;
    struct ixreg *ixreg;

    if (!strcmp(el, "pciid")) {
        if (attr[0] == NULL || strcmp(attr[0], "id")) {
            fprintf(stderr, "(%s:%d) ill formated xml: %d\n",
                    __func__, __LINE__, count_line);
            return;
        }
        pdata->dump->pciid = strtoul(attr[1], NULL, 16);
    }
    if (!strcmp(el, "mm")) {
        unsigned long o, v;
        if (attr[0] == NULL || strcmp(attr[0], "off") ||
            attr[2] == NULL || strcmp(attr[2], "val")) {
            fprintf(stderr, "(%s:%d) ill formated xml: %d\n",
                    __func__, __LINE__, count_line);
            return;
        }
        o = strtoul(attr[1], NULL, 16);
        v = strtoul(attr[3], NULL, 16);
        mmreg = mmreg_new(o, v);
        if (mmreg)
            dump_add_mmreg(pdata->dump, mmreg);
    }
    if (!strcmp(el, "ix")) {
        unsigned long o, v, idx, ida;
        if (attr[0] == NULL || strcmp(attr[0], "indx") ||
            attr[2] == NULL || strcmp(attr[2], "data") ||
            attr[4] == NULL || strcmp(attr[4], "off") ||
            attr[6] == NULL || strcmp(attr[6], "val")) {
            fprintf(stderr, "(%s:%d) ill formated xml: %d\n",
                    __func__, __LINE__, count_line);
            return;
        }
        idx = strtoul(attr[1], NULL, 16);
        ida = strtoul(attr[3], NULL, 16);
        o = strtoul(attr[5], NULL, 16);
        v = strtoul(attr[7], NULL, 16);
        ixreg = ixreg_new(idx, ida, o, v);
        if (ixreg)
            dump_add_ixreg(pdata->dump, ixreg);
    }
}

static void end_func(void *data, const char *el)
{
}

int dump_xml_read(struct dump *dump, const char *filename)
{
    FILE *file; 
    char buf[1024];
    int len;
    int done = 0;
    struct xml_parser pdata;
    XML_Parser parser;
    long end;

    dump_init(dump);
    file = fopen(filename, "r");
    if (file == NULL) {
        fprintf(stderr, "Failed to open: %s\n", filename);
        return 0;
    }
    fseek(file, 0, SEEK_END);
    end = ftell(file);
    fseek(file, 0, SEEK_SET);
    pdata.dump = dump;
    parser = XML_ParserCreate(NULL);
    XML_SetUserData(parser, &pdata);
    XML_SetElementHandler(parser, start_func, end_func);
    do {
        if (fgets(buf, 1024, file) == NULL)
            return -1;
        count_line++;
        len = strlen(buf);
        if (ftell(file) == end) {
            done = 1;
        }
        if (!XML_Parse(parser, buf, len, done)) {
            fprintf(stderr, "at line %d\n", count_line);
            return -1;
        }
    } while (!done);
    XML_ParserFree(parser);
    fclose(file);
    return 0;
}

int dump_xml_write(struct dump *dump, const char *filename)
{
    struct mmreg *mmreg;
    struct ixreg *ixreg;
    FILE *file;

    file = fopen(filename, "w");
    if (file == NULL) {
        fprintf(stderr, "Failed to open: %s\n", filename);
        return -1;
    }
    fprintf(file, "<dump>\n");
    fprintf(file, "<pciid id='0x%08X'/>\n", dump->pciid);
    list_for_each_entry(mmreg, &dump->mmregs, list) {
        fprintf(file, "<mm off='0x%08X' val='0x%08X'/>\n",
                mmreg->offset, mmreg->value);
    }
    list_for_each_entry(ixreg, &dump->ixregs, list) {
        fprintf(file, "<ix indx='0x%08X' data='0x%08X' off='0x%08X' "
                "val='0x%08X'/>\n", ixreg->index, ixreg->data, ixreg->offset,
                ixreg->value);
    }
    fprintf(file, "</dump>\n");
    fclose(file);
    return 0;
}
