/*
 * Copyright 2008 Jerome Glisse <glisse@freedesktop.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and that
 * both that this copyright notice and this permission notice appear in
 * supporting electronic documentation.
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR AUTHORS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <pciaccess.h>
#include "radeon.h"

int radeon_probe(struct radeon *radeon)
{
    struct pci_slot_match match;
    struct pci_device_iterator *iter;
    struct pci_device *device;
    int i = 0;
    int r = 0;

    radeon->bios = (uint8_t*)malloc(256*1024);
    if (radeon->bios == NULL) {
        fprintf(stderr, "Failed to allocate bios copy\n");
        return -1;
    }

    if (pci_system_init()) {
        fprintf(stderr, "Failed to initialize libpciaccess\n");
        return -1;
    }

    match.domain = PCI_MATCH_ANY;
    match.bus = PCI_MATCH_ANY;
    match.dev = PCI_MATCH_ANY;
    match.func = 0;
    match.match_data = 0;
    iter = pci_slot_match_iterator_create(&match);

    while ((device = pci_device_next(iter))) {
        pci_device_probe(device);
        /* ATI vendor id 0x1002 and we are looking for a video
         * card class 0x03 with subclass 0x00 (VGA)
         */
        radeon->pciid = radeon_find_family(device->vendor_id,device->device_id);
        if (radeon->pciid) {
            if (1) {
                printf("Found card %x:%x %s\n",
                       radeon->pciid->vendor,
                       radeon->pciid->device,
                       radeon->pciid->name);
            }
            for (i = 0; i < 6; i++) {
                if (1) {
                    printf("  region: (base: 0x%016X, "
                           "bus: 0x%016X, "
                           "size: %d, is_io: %d)\n",
                           device->regions[i].bus_addr,
                           device->regions[i].base_addr,
                           device->regions[i].size,
                           device->regions[i].is_IO);
                }
                if (device->regions[i].size == 64 * 1024) {
                    radeon->mmio_base = device->regions[i].base_addr;
                    radeon->mmio_size = device->regions[i].size;
                }
            }
            radeon->dev = device;
            break;
        }
    }
    pci_iterator_destroy(iter);
    if (!radeon->dev) {
        fprintf(stderr, "Cannot find radeon device\n");
        return -1;
    }

    if (pci_device_map_range(radeon->dev, radeon->mmio_base, radeon->mmio_size,
                             PCI_DEV_MAP_FLAG_WRITABLE,
                             (void**)&radeon->mmio)) {
        fprintf(stderr, "Mapping mmio failed (are you root ?)\n");
        return -1;
    }
    r = pci_device_read_rom(radeon->dev, radeon->bios);
    if (r) {
        fprintf(stderr, "Cannot read rom %d\n", r);
        return -1;
    }
    return 0;
}

void radeon_close(struct radeon *radeon)
{
    if (pci_device_unmap_range(radeon->dev, radeon->mmio, radeon->mmio_size)) {
        fprintf(stderr, "Unmapping mmio failed\n");
    }
}

u32 radeon_mmio_rd32(struct radeon *radeon, u32 offset)
{
    u32 value;

    if (radeon->mmio == NULL) {
        fprintf(stderr, "(%s:%d) internal error\n", __func__, __LINE__);
        exit(1);
    }
    value = *(u32 * volatile)(radeon->mmio + offset);
    return value;
}

void radeon_mmio_wr32(struct radeon *radeon, u32 offset, u32 value)
{
    if (radeon->mmio == NULL) {
        fprintf(stderr, "(%s:%d) internal error\n", __func__, __LINE__);
        exit(1);
    }
    *(u32 * volatile)(radeon->mmio + offset) = value;
}

int radeon_dump(struct radeon *radeon, struct dump *dump)
{
    struct mmreg *mmreg;
    struct ixreg *ixreg;
    int r;

    if (radeon->pciid == NULL || radeon->pciid->asic == NULL) {
        fprintf(stderr, "(%s:%d) Unsupported chip\n", __func__, __LINE__);
        return -EINVAL;
    }
    list_for_each_entry(mmreg, &dump->mmregs, list) {
        r = radeon->pciid->asic->mmrd32(radeon, mmreg);
        if (r) {
            return r;
        }
    }
    list_for_each_entry(ixreg, &dump->ixregs, list) {
        r = radeon->pciid->asic->ixrd32(radeon, ixreg);
        if (r) {
            return r;
        }
    }
    return 0;
}
